import 'package:extensions_pack/extensions_pack.dart';
import 'package:flutter/material.dart';
import 'package:widgets_pack/widgets/widgets.dart';

sealed class TableAction<M extends Object> {
  final String label;
  final Widget icon;
  final void Function(List<M>)? action;
  final bool Function(List<M>)? visible;

  const TableAction({
    required this.label,
    required this.icon,
    this.action,
    this.visible,
  });
}

class TableSingleAction<M extends Object> extends TableAction<M> {
  TableSingleAction({
    required String label,
    required Widget icon,
    void Function(M)? action,
    bool Function(M)? visible,
  }) : super(
          label: label,
          icon: icon,
          action: action != null ? (items) => action(items.first) : null,
          visible: (items) => visible?.call(items.first) ?? true,
        );
}

class TableMultiAction<M extends Object> extends TableAction<M> {
  TableMultiAction({
    required String label,
    required Widget icon,
    void Function(List<M>)? action,
    bool Function(List<M>)? visible,
  }) : super(
          label: label,
          icon: icon,
          action: action,
          visible: visible,
        );
}

class TableActionDivider<M extends Object> extends TableAction<M> {
  TableActionDivider()
      : super(
          label: '',
          icon: const SizedBox(),
          action: (_) {},
        );
}

class AppTableActionsRow<M extends Object> extends StatelessWidget {
  final List<M> items;
  final List<TableAction<M>> actions;
  final VoidCallback onClearAll;

  const AppTableActionsRow({
    required this.items,
    required this.actions,
    required this.onClearAll,
    super.key,
  });

  Widget? _buildAction(TableAction<M> action) {
    if (action is TableActionDivider<M>) {
      return const VerticalDivider(
        indent: 8,
        endIndent: 8,
      );
    }

    final enabled = action.visible?.call(items) ?? true;

    if (enabled) {
      return AppButton.icon(
        onPressed: action.action != null ? () => action.action!(items) : null,
        tooltip: action.label,
        icon: action.icon,
      );
    }
  }

  @override
  Widget build(BuildContext context) {
    Iterable<Widget> actionsWidgets = (items.length == 1)
        ? actions.map(_buildAction).whereNotNull()
        : actions.whereNot((a) => a is TableSingleAction<M>).map(_buildAction).whereNotNull();

    if (actionsWidgets.first is VerticalDivider) {
      actionsWidgets = actionsWidgets.skip(1);
    }

    return Container(
      height: 46,
      padding: const EdgeInsets.symmetric(horizontal: 16),
      decoration: BoxDecoration(
        color: context.colorTheme.onSecondary,
        borderRadius: BorderRadius.circular(25),
      ),
      child: Row(
        children: <Widget>[
          AppButton.icon(
            onPressed: onClearAll,
            icon: const Icon(Icons.close),
          ),
          TitleSmall(
            '${items.length} selected',
          ),
          const VerticalDivider(
            indent: 8,
            endIndent: 8,
          ),
          ...actionsWidgets,
        ].addSpacingBetween(),
      ),
    );
  }
}
